#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

// This kernel evenly distributes workloads across threads for the matrix multiplication (linear layer)
// and the pooling stage. For each sample b, all threads in the block cooperatively load the input vector
// into shared memory, compute a portion of the output features, and then perform max pooling in an evenly
// distributed manner using only the necessary number of threads for pooling. This avoids bottlenecks when
// the number of pooling segments is less than the block size.

__global__ void evenly_distributed_kernel(
    const float* __restrict__ x,
    const float* __restrict__ weight,
    const float* __restrict__ bias,
    float* __restrict__ output,
    int B,           // batch size
    int inF,         // number of input features
    int outF,        // number of output features
    int kernel_size,
    float scale_factor
) {
    // Each block processes one sample in the batch
    int b = blockIdx.x;
    if (b >= B) return;

    // Shared memory layout:
    // [0, inF)         -> shared copy of input vector x[b]
    // [inF, inF+outF)  -> local_out array (linear transformation result)
    // [inF+outF, ...)  -> temporary storage for pooling reduction (size = blockDim.x)
    extern __shared__ float smem[];
    float* shared_x = smem;                    // size: inF
    float* local_out = shared_x + inF;         // size: outF
    float* pool_shared = local_out + outF;     // size: blockDim.x (for pooling reduction)

    // 1. Cooperative load of input vector x[b] into shared memory
    for (int i = threadIdx.x; i < inF; i += blockDim.x) {
        shared_x[i] = x[b * inF + i];
    }
    __syncthreads();

    // 2. Compute the linear transformation for each output feature j:
    //    local_out[j] = bias[j] + dot(x[b], weight[j])
    for (int j = threadIdx.x; j < outF; j += blockDim.x) {
        float sum_val = bias[j];
        // Use shared_x for input vector
        for (int i = 0; i < inF; i++) {
            sum_val = __fmaf_rn(shared_x[i], weight[j * inF + i], sum_val);
        }
        local_out[j] = sum_val;
    }
    __syncthreads();

    // 3. Max pooling over segments of size 'kernel_size'
    //    Each pooling segment covers kernel_size consecutive output features
    int pooled_len = (outF >= kernel_size) ? (1 + (outF - kernel_size) / kernel_size) : 0;

    // Distribute pooling work evenly:
    // Use only as many threads as needed if pooled_len is smaller than blockDim.x
    int pool_threads = (pooled_len < blockDim.x) ? pooled_len : blockDim.x;
    float pool_local = 0.0f;

    if (pooled_len > 0 && threadIdx.x < pool_threads) {
        // Each active thread processes multiple segments in a grid-stride loop
        for (int seg = threadIdx.x; seg < pooled_len; seg += pool_threads) {
            int start = seg * kernel_size;
            float seg_max = local_out[start];
            for (int k = 1; k < kernel_size; k++) {
                seg_max = fmaxf(seg_max, local_out[start + k]);
            }
            pool_local += seg_max;
        }
    }

    // 4. Reduction of pooling results among the active pool threads
    if (threadIdx.x < pool_threads) {
        pool_shared[threadIdx.x] = pool_local;
    }
    __syncthreads();

    // Simple tree reduction in shared memory
    for (int stride = pool_threads / 2; stride > 0; stride /= 2) {
        if (threadIdx.x < stride && threadIdx.x < pool_threads) {
            pool_shared[threadIdx.x] += pool_shared[threadIdx.x + stride];
        }
        __syncthreads();
    }

    // 5. Write the final result for sample b
    if (threadIdx.x == 0) {
        output[b] = pool_shared[0] * scale_factor;
    }
}

// Forward function callable from Python
at::Tensor forward(
    at::Tensor x,
    int64_t kernel_size,
    double scale_factor,
    at::Tensor weight,
    at::Tensor bias
) {
    TORCH_CHECK(x.is_cuda(), "x must be a CUDA tensor");
    TORCH_CHECK(weight.is_cuda(), "weight must be a CUDA tensor");
    TORCH_CHECK(bias.is_cuda(), "bias must be a CUDA tensor");
    TORCH_CHECK(x.dim() == 2, "x must have shape (batch_size, in_features)");
    TORCH_CHECK(weight.dim() == 2, "weight must have shape (out_features, in_features)");
    TORCH_CHECK(bias.dim() == 1, "bias must have shape (out_features)");

    int B = x.size(0);
    int inF = x.size(1);
    int outF = weight.size(0);

    auto out = torch::empty({B}, x.options());

    // Set block size and calculate shared memory usage:
    const int threads = 256;
    size_t sharedMemSize = (inF + outF + threads) * sizeof(float);

    // Launch one block per sample in the batch
    evenly_distributed_kernel<<<B, threads, sharedMemSize>>>(
        x.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.data_ptr<float>(),
        out.data_ptr<float>(),
        B,
        inF,
        outF,
        static_cast<int>(kernel_size),
        static_cast<float>(scale_factor)
    );

    return out;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Evenly distributed CUDA forward kernel");
}
